
/* Chrysalide - Outil d'analyse de fichiers binaires
 * maxcommon.c - détermination de la plus grand occurrence au sein d'un ensemble d'éléments
 *
 * Copyright (C) 2023 Cyrille Bagard
 *
 *  This file is part of Chrysalide.
 *
 *  Chrysalide is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Chrysalide is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 */


#include "maxcommon.h"


#include <assert.h>
#include <malloc.h>


#include "../item-int.h"
#include "../exprs/literal.h"
#include "../../../glibext/comparison-int.h"



/* ---------------------- INTRODUCTION D'UNE NOUVELLE FONCTION ---------------------- */


/* Initialise la classe des repérages de plus grande occurrence. */
static void g_scan_maxcommon_function_class_init(GScanMaxcommonFunctionClass *);

/* Initialise une instance de repérage d'occurrence maximake. */
static void g_scan_maxcommon_function_init(GScanMaxcommonFunction *);

/* Supprime toutes les références externes. */
static void g_scan_maxcommon_function_dispose(GScanMaxcommonFunction *);

/* Procède à la libération totale de la mémoire. */
static void g_scan_maxcommon_function_finalize(GScanMaxcommonFunction *);



/* --------------------- IMPLEMENTATION DES FONCTIONS DE CLASSE --------------------- */


/* Indique le nom associé à une expression d'évaluation. */
static char *g_scan_maxcommon_function_get_name(const GScanMaxcommonFunction *);

/* Réduit une expression à une forme plus simple. */
static bool g_scan_maxcommon_function_run_call(GScanMaxcommonFunction *, GScanExpression **, size_t, GScanContext *, GScanScope *, GObject **);



/* ---------------------------------------------------------------------------------- */
/*                        INTRODUCTION D'UNE NOUVELLE FONCTION                        */
/* ---------------------------------------------------------------------------------- */


/* Indique le type défini pour un décompte d'élément le plus représenté dans une série. */
G_DEFINE_TYPE(GScanMaxcommonFunction, g_scan_maxcommon_function, G_TYPE_SCAN_REGISTERED_ITEM);


/******************************************************************************
*                                                                             *
*  Paramètres  : klass = classe à initialiser.                                *
*                                                                             *
*  Description : Initialise la classe des repérages de plus grande occurrence.*
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_scan_maxcommon_function_class_init(GScanMaxcommonFunctionClass *klass)
{
    GObjectClass *object;                   /* Autre version de la classe  */
    GScanRegisteredItemClass *registered;   /* Version de classe parente   */

    object = G_OBJECT_CLASS(klass);

    object->dispose = (GObjectFinalizeFunc/* ! */)g_scan_maxcommon_function_dispose;
    object->finalize = (GObjectFinalizeFunc)g_scan_maxcommon_function_finalize;

    registered = G_SCAN_REGISTERED_ITEM_CLASS(klass);

    registered->get_name = (get_registered_item_name_fc)g_scan_maxcommon_function_get_name;
    registered->run_call = (run_registered_item_call_fc)g_scan_maxcommon_function_run_call;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : func = instance à initialiser.                               *
*                                                                             *
*  Description : Initialise une instance de repérage d'occurrence maximake.   *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_scan_maxcommon_function_init(GScanMaxcommonFunction *func)
{

}


/******************************************************************************
*                                                                             *
*  Paramètres  : func = instance d'objet GLib à traiter.                      *
*                                                                             *
*  Description : Supprime toutes les références externes.                     *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_scan_maxcommon_function_dispose(GScanMaxcommonFunction *func)
{
    G_OBJECT_CLASS(g_scan_maxcommon_function_parent_class)->dispose(G_OBJECT(func));

}


/******************************************************************************
*                                                                             *
*  Paramètres  : func = instance d'objet GLib à traiter.                      *
*                                                                             *
*  Description : Procède à la libération totale de la mémoire.                *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_scan_maxcommon_function_finalize(GScanMaxcommonFunction *func)
{
    G_OBJECT_CLASS(g_scan_maxcommon_function_parent_class)->finalize(G_OBJECT(func));

}


/******************************************************************************
*                                                                             *
*  Paramètres  : -                                                            *
*                                                                             *
*  Description : Constitue une fonction de calcul de plus grande occurrence.  *
*                                                                             *
*  Retour      : Fonction mise en place.                                      *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

GScanRegisteredItem *g_scan_maxcommon_function_new(void)
{
    GScanRegisteredItem *result;            /* Structure à retourner       */

    result = g_object_new(G_TYPE_SCAN_MAXCOMMON_FUNCTION, NULL);

    return result;

}



/* ---------------------------------------------------------------------------------- */
/*                       IMPLEMENTATION DES FONCTIONS DE CLASSE                       */
/* ---------------------------------------------------------------------------------- */


/******************************************************************************
*                                                                             *
*  Paramètres  : item = élément d'appel à consulter.                          *
*                                                                             *
*  Description : Indique le nom associé à une expression d'évaluation.        *
*                                                                             *
*  Retour      : Désignation humaine de l'expression d'évaluation.            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static char *g_scan_maxcommon_function_get_name(const GScanMaxcommonFunction *item)
{
    char *result;                           /* Désignation à retourner     */

    result = strdup("maxcommon");

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : item  = élément d'appel à consulter.                         *
*                args  = liste d'éventuels arguments fournis.                 *
*                count = taille de cette liste.                               *
*                ctx   = contexte de suivi de l'analyse courante.             *
*                scope = portée courante des variables locales.               *
*                out   = zone d'enregistrement de la résolution opérée. [OUT] *
*                                                                             *
*  Description : Réduit une expression à une forme plus simple.               *
*                                                                             *
*  Retour      : Réduction correspondante, expression déjà réduite, ou NULL.  *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static bool g_scan_maxcommon_function_run_call(GScanMaxcommonFunction *item, GScanExpression **args, size_t count, GScanContext *ctx, GScanScope *scope, GObject **out)
{
    bool result;                            /* Bilan à retourner           */
    size_t used;                            /* Prochain emplacement libre  */
    GScanExpression **collected;            /* Représentants de groupes    */
    size_t *scores;                         /* Taille de ces groupes       */
    size_t i;                               /* Boucle de parcours #1       */
    size_t k;                               /* Boucle de parcours #2       */
    bool status;                            /* Bilan de la comparaison     */
    bool equal;                             /* Egalité établie ?           */
    size_t arg0_count;                      /* Taille de l'argument unique */
    GScanExpression *arg0_item;             /*  Elément de cet argument    */
    size_t best;                            /* Meilleur score identifié    */

    result = (count > 0);
    if (!result) goto exit;

    used = 0;

    /* Si la série à étudier est directement fournie */
    if (count > 1)
    {
        collected = malloc(count * sizeof(GScanExpression *));
        scores = malloc(count * sizeof(size_t));

        for (i = 0; i < count; i++)
        {
            for (k = 0; k < used; k++)
            {
                status = g_comparable_item_compare_rich(G_COMPARABLE_ITEM(args[i]),
                                                        G_COMPARABLE_ITEM(collected[k]),
                                                        RCO_EQ, &equal);

                if (status && equal)
                    break;

            }

            if (k < used)
                scores[k]++;

            else
            {
                collected[used] = args[i];
                g_object_ref(G_OBJECT(args[i]));
                scores[used] = 1;

                used++;

            }

        }

    }

    /* Sinon on considère que l'arguement unique porte la liste (idéalement) */
    else
    {
        if (G_IS_SCAN_LITERAL_EXPRESSION(args[0]) || !g_scan_expression_handle_set_features(args[0]))
        {
            best = 1;
            goto quick_unique;
        }

#ifndef NDEBUG
        g_scan_expression_count_items(args[0], ctx, &arg0_count);
#else
        status = bool g_scan_expression_count_items(args[0], ctx, &arg0_count);
        assert(status);
#endif

        collected = malloc(arg0_count * sizeof(GScanExpression *));
        scores = malloc(arg0_count * sizeof(size_t));

        for (i = 0; i < arg0_count; i++)
        {
#ifndef NDEBUG
            g_scan_expression_get_item(args[0], i, ctx, &arg0_item);
#else
            status = g_scan_expression_get_item(args[0], i, ctx, &arg0_item);
            assert(status);
#endif

            for (k = 0; k < used; k++)
            {
                status = g_comparable_item_compare_rich(G_COMPARABLE_ITEM(arg0_item),
                                                        G_COMPARABLE_ITEM(collected[k]),
                                                        RCO_EQ, &equal);

                if (status && equal)
                    break;

            }

            if (k < used)
            {
                g_object_unref(G_OBJECT(arg0_item));
                scores[k]++;
            }

            else
            {
                collected[used] = arg0_item;
                scores[used] = 1;

                used++;

            }

        }

    }

    /* Analyse des résultats */

    best = 0;

    for (i = 0; i < used; i++)
        if (scores[i] > best)
            best = scores[i];

    for (i = 0; i < used; i++)
        g_object_unref(G_OBJECT(collected[i]));

    free(collected);
    free(scores);

 quick_unique:

    assert(best > 0);

    *out = G_OBJECT(g_scan_literal_expression_new(LVT_UNSIGNED_INTEGER, (unsigned long long []){ best }));

 exit:

    return result;

}
